# librsa
# by Benjamin Randazzo - 2012
# based on Sybren Stuvel, Marloes de Boer and Ivo Tamboer work (2009)

#!/bin/python

import random, time, math, zlib, base64, types
from cPickle import dumps, loads

def genkey(name):
	print ". . .",	
	(n, e, p, q) = genPubKey()
	time.sleep(3)

	print ". . .",
	d = genPrivKey(e, p, q)

	# write public key
	f = open(name+".pub", 'w')
	f.write('----- BEGIN PUBLIC KEY -----\n'+
		base64.encodestring("(n = "+str(n)+", e = "+str(e)+")")+
		'-----  END PUBLIC KEY  -----\n')
	f.close()

	# write private key
	f = open(name+".priv", 'w')
	f.write('----- BEGIN PRIVATE KEY -----\n'+
		base64.encodestring("(n = "+str(n)+", d = "+str(d)+")")+
		'-----  END PRIVATE KEY  -----\n')
	f.close()

def genPubKey():
	p = random.random()
	p *= 10**8
	p = nextPrime(int(p))

	q = random.random()
	q *= 10**8
	q = nextPrime(int(q))

	n = p*q
	e = 101
	print "+ + +",
	return (n, e, p, q)

def genPrivKey(e, p, q):
	m = (p-1)*(q-1)
	print "+ + +",
	return mul_inv(e, m)

def nextPrime(n):
	isPrime = False
	while (not isPrime):
		n = n+1
		limit = int(math.sqrt(n)+1)
		i = 2
		while (i < limit and n%i != 0):
			i = i+1
		if i == limit:
			isPrime = True
	return n

def bezout(a, b):
	if a == 0 and b == 0: return (0, 0, 0)
	if b == 0: return (a/abs(a), 0, abs(a))
	(u, v, p) = bezout(b, a%b)
	return (v, (u - v*(a/b)), p)

def mul_inv(x, m):
	(u, _, p) = bezout(x, m)
	if p == 1: return u%abs(m)

def int2bytes(n):
	if not (type(n) is types.LongType or type(n) is types.IntType):
	        raise TypeError("You must pass a long or an int")
	
	string = ""
	while n > 0:
		string = "%s%s" % (chr(n & 0xFF), string)
		n /= 256
	return string

def bytes2int(bytes):
	integer = 0
	for byte in bytes:
		integer *= 256
		if type(byte) is types.StringType: byte = ord(byte)
		integer += byte
	return integer

def chopstring(message, key, n, funcref):
	msglen = len(message)
	mbits = msglen*8
	nbits = int(math.floor(math.log(n, 2)))
	nbytes = nbits/8
	blocks = msglen/nbytes

	if msglen%nbytes > 0:
		blocks += 1

	cipher = []
	for bindex in range(blocks):
		offset = bindex*nbytes
		block = message[offset:offset+nbytes]
		value = bytes2int(block)
		cipher.append(funcref(value, key, n))
	return picklechops(cipher)

def picklechops(chops):
	value = zlib.compress(dumps(chops))
	encoded = base64.encodestring(value)
	return encoded.strip()

def unpicklechops(string):
	return loads(zlib.decompress(base64.decodestring(string)))

def fast_exponentiation(a, p, n):
	result = a%n
	remainders = []
	while p != 1:
		remainders.append(p & 1)
		p = p >> 1
	while remainders:
		rem = remainders.pop()
		result = ((a ** rem) * result ** 2) % n
	return result

def getCiphertext(message, e, n):
	if type(message) is types.IntType:
		return getCiphertext(long(message), e, n)

	if message > 0 and math.floor(math.log(message, 2)) > math.floor(math.log(n, 2)):
		raise OverflowError("The message is too long")
	return fast_exponentiation(message, e, n)

def getPubKey(filename):
	f = open(filename, 'r')
	f.readline()
	pubKey = f.readline()
	f.close()
	return extractPubKey(pubKey)

def extractPubKey(pubKey):
	charset = '0123456789'

	pub = base64.decodestring(pubKey)
	n = e = ''
	modulus = True
	for i in pub:
		if (i == ','):
			modulus = False
		if (modulus == True and i in charset):
			n += i
		elif (modulus == False and i in charset):
			e += i
	return (n, e)

def getMessage(filename):
	f = open(filename, 'r')
	m = f.read()
	f.close()
	return m

def encrypt(n, e, filename):
	message = getMessage(filename)

	f = open(filename+'.asc', 'w')
	f.write('----- BEGIN MESSAGE -----\n'+
		chopstring(message, int(e), int(n), getCiphertext)+'\n'+
		'-----  END MESSAGE  -----\n')
	f.close()
	
def getPrivKey(filename):
	f = open(filename, 'r')
	f.readline()
	privKey = f.readline()
	f.close()
	return extractPrivKey(privKey)

def extractPrivKey(privKey):
	charset = '0123456789'

	priv = base64.decodestring(privKey)
	n = d = ''
	modulus = True
	for i in priv:
		if (i == ','):
			modulus = False
		if (modulus == True and i in charset):
			n += i
		elif (modulus == False and i in charset):
			d += i
	return (n, d)

def getUnknownMessage(filename):
	f = open(filename, 'r')
	m = f.read()
	f.close()
	m = m.rstrip('\n')
	m = m[len('----- BEGIN MESSAGE -----')+1:-(len('-----  END MESSAGE  -----')+1)]
	return m

def getPlaintext(message, d ,n):
	 return getCiphertext(message, d, n)

def gluechops(chops, key, n, funcref):
	message = ""
	chops = unpicklechops(chops)

	for cpart in chops:
		mpart = funcref(cpart, key, n)
		message += int2bytes(mpart)
	return message

def decrypt(n, d, filename):
	message = getUnknownMessage(filename)

	f = open(filename.rstrip('.asc'), 'w')
	f.write(gluechops(message, int(d), int(n), getPlaintext))
	f.close()
